function [K] = h_cluster_kernel(Xl, Xu, method, varargin)
% Hierarchical cluster kernel
% Returns the (graph-based) hierarchical cluster kernel ([g]HCK)
%
% Xl: labeled points
% Xu: unlabeled points
% method: hierarchical clustering method: 'single', 'complete', 'average'
% Optional parameters:
%   K: number of nearest neighbors
%   p_type: 'graph' OR 'euclidean' distances used for increment when connecting the components
%   
% (c) 2011 Zalan Bodo

% Setting parameters
K = 0;
p_type = 'graph'; % 'euclidean'
for i=1:size(varargin, 2)
    switch i
        case 1
            K = varargin{i};
        case 2
            p_type = varargin{i};
    end
end

% Some auxiliary variables
X = [Xl; Xu];
N = size(X,1);
Nl = size(Xl,1);
Nu = N - Nl;
dim = size(X,2);

D = squareform(pdist(X, 'euclidean'));
D_eucl = D; 

if (K > 0),
    
    % Constructing the neighborhood graph
    disp('Constructing neighborhood graph...'); 

    [tmp, ind] = sort(D); 
    for i=1:N
      D(i, ind((2+K):end,i)) = Inf; 
    end
    % Symmetrize distance matrix (numerical errors)
    D = min(D, D'); 

    % Computing shortest path distances using Floyd-Washall's algorithm
    tic; 
    for k=1:N
         D = min(D, repmat(D(:, k), [1 N]) + repmat(D(k, :), [N 1])); 
         if (rem(k, 20) == 0)
              disp([' Iteration: ' num2str(k) 'Estimated time to completion: ' num2str((N-k)*toc/k/60) ' minutes']); 
         end
    end

    % Modifying D to be a connected graph, i.e. connecting components (in case)
    disp('Connecting components...');
    
    L = 0; dm = 0;
    i = []; j = [];
    [i, j] = find(D == Inf);
  
    if ~isempty(i)
        
        if p_type == 'graph'
            dmax = max(setdiff(reshape(D, N*N, 1), Inf));
            dmin = min(setdiff(reshape(D, N*N, 1), 0));
            gmax = dmax;
        else
            if p_type == 'euclidean'
                dmax = max(setdiff(reshape(D_eucl, N*N, 1), Inf));
                dmin = min(setdiff(reshape(D_eucl, N*N, 1), 0));
                gmax = max(setdiff(reshape(D, N*N, 1), Inf));
            end
        end
        
        L = gmax + (dmin/dmax);

        k = 1;
        while ~isempty(i)
            disp(['Round ' k]); k = k+1;
            
            [min_eucl, ii] = min(D_eucl(sub2ind(size(D), i, j)));
            D(i(ii), j(ii)) = L; D(j(ii), i(ii)) = L; % for symmetry
            
            % Calculating new distances
            for m = 1:size(i,1)
                D(i(m),j(m)) = min(D(i(m),i(ii))+L+D(j(ii),j(m)), D(i(m),j(ii))+L+D(i(ii),j(m)));
                D(j(m),i(m)) = D(i(m),j(m));
            end

            [i, j] = find(D == Inf);
            L = L + (dmin/dmax);
        end % while
        
        dm = dmin/dmax;
    end %if
end %if

% Performing hierarchical clustering
disp('Hierarchical clustering...');
if (K > 0)
    disp('[Graph distances]');
    UD = hcluster_all(D, 1, method);
else
    disp('[Euclidean distances]');
    UD = hcluster_all(X, 0, method);
end;

% Computing the kernel
J = eye(N) - (1/N)*ones(N,N);
K = -0.5*J*UD*J;
